import asyncio
import uuid
import pickle

from functools import partial

from .base import create_session_factory
from .cookie_session import SecureCookie, SessionBackendStore


__all__ = [
    'RedisBackend',
    'RedisSessionFactory',
    ]


_loads = partial(pickle.loads)
_dumps = partial(pickle.dumps, protocol=pickle.HIGHEST_PROTOCOL)


def RedisSessionFactory(redis, secret_key, cookie_name, *,
                        loads=_loads, dumps=_dumps, key_prefix='session:',
                        session_max_age=None, loop=None, **kwargs):
    cookie_store = SecureCookie(secret_key, cookie_name,
                                session_max_age=session_max_age,
                                **kwargs)
    backend = RedisBackend(redis, loads=loads, dumps=dumps,
                           session_max_age=session_max_age)
    return create_session_factory(session_id_store=cookie_store,
                                  backend_store=backend,
                                  loop=loop)


class RedisBackend(SessionBackendStore):
    """Redis session store backend.
    """

    def __init__(self, redis, *, loads=_loads, dumps=_dumps,
                 key_prefix='session:', session_max_age=None):
        self._redis = redis
        self._loads = loads
        self._dumps = dumps
        self._key_prefix = key_prefix
        self.session_max_age = session_max_age

    @asyncio.coroutine
    def load_session_data(self, cookie_value):
        """Loads session data pointed by sid stored in cookies.
        """
        sid = cookie_value
        key = self._make_key(sid)
        packed = yield from self._redis.get(key)
        if packed is not None:
            try:
                data = self._loads(packed)
            except (TypeError, ValueError):
                pass
            else:
                return data, sid
        return None, None

    @asyncio.coroutine
    def save_session_data(self, session):
        """Saves session data in redis and returns sid.
        """
        if session.new:
            sid = self.new_sid()
        else:
            sid = session.identity
        key = self._make_key(sid)
        if not session and sid:
            yield from self._redis.delete([key])
            return None
        data = self._dumps(dict(session))
        if self.session_max_age is not None:
            yield from self._redis.setex(key, self.session_max_age, data)
        else:
            yield from self._redis.set(key, data)
        return sid

    def new_sid(self):
        """Returns new sid.
        """
        return uuid.uuid4().hex

    def _make_key(self, sid):
        return '{prefix}{sid}'.format(
            prefix=self._key_prefix, sid=sid).encode('utf-8')



"""Cookie based sessions."""
import asyncio
import time
import hmac
import hashlib

from .base import create_session_factory
from .interface import SessionIdStore, SessionBackendStore


__all__ = [
    'CookieSessionFactory',
    'SecureCookie',
    ]


def CookieSessionFactory(loads, dumps, secret_key, cookie_name,
                         session_max_age=None,
                         domain=None, max_age=None, path=None,
                         secure=None, httponly=None,
                         *, loop=None):
    sid_store = SecureCookie(secret_key, cookie_name,
                             session_max_age=session_max_age,
                             domain=domain, max_age=max_age,
                             path=path, secure=secure,
                             httponly=httponly)
    backend = ClientSideBackend(loads, dumps)
    return create_session_factory(session_id_store=sid_store,
                                  backend_store=backend,
                                  loop=loop)


class ClientSideBackend(SessionBackendStore):
    """Client-side backend.

    Stores session object in session_id.
    """

    def __init__(self, loads, dumps):
        self._loads = loads
        self._dumps = dumps

    @asyncio.coroutine
    def load_session_data(self, cookie_value):
        """Load session data from decoded and verified cookie value.

        Returns session dict or None and empty string as session id
        """
        try:
            return self._loads(cookie_value), ''
        except (TypeError, ValueError):
            # TODO: log warning
            pass
        return None, None

    @asyncio.coroutine
    def save_session_data(self, session):
        """Save session and return bytes value to be stored in cookie.
        """
        if not session:
            return
        return self._dumps(dict(session))


class SecureCookie(SessionIdStore):
    """Cookie-based session factory.

    Stores session in cookies.
    """

    def __init__(self, secret_key, cookie_name,
                 session_max_age=None,
                 domain=None, max_age=None, path=None,
                 secure=None, httponly=None):
        if isinstance(secret_key, str):
            secret_key = secret_key.encode('utf-8')
        self._secret_key = secret_key
        self._cookie_name = cookie_name
        self._cookie_params = dict(domain=domain,
                                   max_age=max_age,
                                   path=path,
                                   secure=secure,
                                   httponly=httponly)
        self.session_max_age = session_max_age

    def get_session_id(self, request):
        name = self._cookie_name
        raw_value = request.cookies.get(name)
        return self._decode_cookie(raw_value)

    def put_session_id(self, request, cookie_value):
        if cookie_value is None:
            request.response.del_cookie(self._cookie_name)
        else:
            raw_value = self._encode_cookie(cookie_value)
            request.response.set_cookie(self._cookie_name, raw_value,
                                        **self._cookie_params)

    def _encode_cookie(self, value):
        """Encode and sign cookie value.

        value argument must be str instance.
        """
        assert isinstance(value, str)
        name = self._cookie_name
        timestamp = str(int(time.time()))
        singature = self._get_signature(name, value, timestamp)
        return '|'.join((value, timestamp, singature))

    def _decode_cookie(self, value):
        """Decode and verify cookie value.

        value argument must be str.
        Returns decoded bytes value of cookie
        or None if value could not be decoded or verified.
        """
        if not value:
            return None
        parts = value.split('|')
        if len(parts) != 3:
            return None
        name = self._cookie_name
        value, timestamp, sign = parts

        if self.session_max_age is not None:
            if int(timestamp) < int(time.time()) - self.session_max_age:
                return None

        expected_sign = self._get_signature(name, value, timestamp)
        if not hmac.compare_digest(expected_sign, sign):
            # TODO: log warning
            return None
        return value

    def _get_signature(self, *parts):
        sign = hmac.new(self._secret_key, digestmod=hashlib.sha1)
        sign.update(('|'.join(parts)).encode('utf-8'))
        return sign.hexdigest()



from itertools import chain
from collections import OrderedDict, abc


class MultiDict(abc.Mapping):
    """Read-only ordered dictionary that can hava multiple values for each key.

    This type of MultiDict must be used for request headers and query args.
    """

    def __init__(self, *args, **kwargs):
        if len(args) > 1:
            raise TypeError("MultiDict takes at most 2 positional "
                            "arguments ({} given)".format(len(args) + 1))
        self._items = OrderedDict()
        if args:
            if hasattr(args[0], 'items'):
                args = list(args[0].items())
            else:
                args = list(args[0])
        for key, value in chain(args, kwargs.items()):
            if key in self._items:
                self._items[key].append(value)
            else:
                self._items[key] = [value]

    # multidict interface suitable for wtforms #

    def getall(self, key):
        """Returns all values stored at key as a tuple.

        Raises KeyError if key doesn't exist.
        """
        return tuple(self._items[key])

    def getone(self, key):
        """Return first value stored at key.
        """
        return self._items[key][0]

    # extra methods #

    def copy(self):
        """Returns a copy itself.
        """
        cls = self.__class__
        return cls(self.items(getall=True))

    # Mapping interface #

    def __getitem__(self, key):
        return self._items[key][0]

    def __iter__(self):
        return iter(self._items)

    def __len__(self):
        return len(self._items)

    def items(self, *, getall=False):
        return _ItemsView(self._items, getall=getall)

    def values(self, *, getall=False):
        return _ValuesView(self._items, getall=getall)

    def __eq__(self, other):
        if not isinstance(other, abc.Mapping):
            return NotImplemented
        if isinstance(other, MultiDict):
            return self._items == other._items
        return dict(self.items()) == dict(other.items())

    def __repr__(self):
        return '<{} {!r}>'.format(self.__class__.__name__, self._items)


class MutableMultiDict(abc.MutableMapping, MultiDict):
    """An ordered dictionary that can have multiple values for each key.
    """

    def getall(self, key):
        """Returns all values stored at key as list.

        Raises KeyError if key doesn't exist.
        """
        return list(self._items[key])

    def add(self, key, value):
        """Adds value to a key.
        """
        if key in self._items:
            self._items[key].append(value)
        else:
            self._items[key] = [value]

    def extend(self, *args, **kwargs):
        """Extends current MutableMultiDict with more values.

        This method must be used instead of update.
        """
        if len(args) > 1:
            raise TypeError("extend takes at most 2 positional arguments"
                            " ({} given)".format(len(args) + 1))
        if args:
            if isinstance(args[0], MultiDict):
                items = args[0].items(getall=True)
            elif hasattr(args[0], 'items'):
                items = args[0].items()
            else:
                items = args[0]
        else:
            items = []
        for key, value in chain(items, kwargs.items()):
            self.add(key, value)

    def clear(self):
        """Remove all items from MutableMultiDict
        """
        self._items.clear()

    # MutableMapping interface #

    def __setitem__(self, key, value):
        self._items[key] = [value]

    def __delitem__(self, key):
        del self._items[key]

    def pop(self, key, default=None):
        """Method not allowed."""
        raise NotImplementedError

    def popitem(self):
        """Method not allowed."""
        raise NotImplementedError

    def update(self, *args, **kw):
        """Method not allowed."""
        raise NotImplementedError("Use extend method instead")


class _ItemsView(abc.ItemsView):

    def __init__(self, mapping, *, getall=False):
        super().__init__(mapping)
        self._getall = getall

    def __contains__(self, item):
        key, value = item
        try:
            values = self._mapping[key]
        except KeyError:
            return False
        else:
            if self._getall:
                return value in values
            else:
                return value == values[0]

    def __iter__(self):
        for key, values in self._mapping.items():
            if self._getall:
                for value in values:
                    yield key, value
            else:
                yield key, values[0]


class _ValuesView(abc.KeysView):

    def __init__(self, mapping, *, getall=False):
        super().__init__(mapping)
        self._getall = getall

    def __contains__(self, value):
        for values in self._mapping.values():
            if self._getall and value in values:
                return True
            elif value == values[0]:
                return True
        return False

    def __iter__(self):
        for values in self._mapping.values():
            if self._getall:
                yield from iter(values)
            else:
                yield values[0]



import asyncio
from collections import MutableMapping

from .interface import SessionIdStore, SessionBackendStore

__all__ = ['Session', 'create_session_factory']


class Session(MutableMapping):
    """Session dict-like object.
    """

    def __init__(self, data=None, identity=None):
        self._changed = False
        self._mapping = {}
        self._identity = identity
        if data is not None:
            self._mapping.update(data)

    def __repr__(self):
        return '<{} [new:{}, changed:{}] {!r}>'.format(
            self.__class__.__name__, self.new, self._changed,
            self._mapping)

    @property
    def new(self):
        return self._identity is None

    @property
    def identity(self):
        return self._identity

    def changed(self):
        self._changed = True

    def invalidate(self):
        self._changed = True
        self._mapping = {}

    def __len__(self):
        return len(self._mapping)

    def __iter__(self):
        return iter(self._mapping)

    def __contains__(self, key):
        return key in self._mapping

    def __getitem__(self, key):
        return self._mapping[key]

    def __setitem__(self, key, value):
        self._mapping[key] = value
        self._changed = True

    def __delitem__(self, key):
        del self._mapping[key]
        self._changed = True


class _SessionFactory:
    """Session factory.
    """

    def __init__(self, *, session_id_store, backend_store, loop=None):
        if loop is None:
            loop = asyncio.get_event_loop()
        self._loop = loop
        assert isinstance(session_id_store, SessionIdStore), session_id_store
        assert isinstance(backend_store, SessionBackendStore), backend_store
        self._sid_store = session_id_store
        self._backend = backend_store

    def __call__(self, request, fut):
        """Instantiate Session object.
        """
        return asyncio.Task(self._load(request, fut), loop=self._loop)

    @asyncio.coroutine
    def _load(self, request, fut):
        """Load or create new session.
        """
        try:
            session_id = self._sid_store.get_session_id(request)
            if session_id is None:
                sess = Session()
            else:
                data, ident = yield from self._backend.load_session_data(
                    session_id)
                sess = Session(data, identity=ident)
        except Exception as exc:
            fut.set_exception(exc)
        else:
            fut.set_result(sess)
            # FIXME: the next line is a subject to the following issue:
            #   yield from request.session  # (response callback added)
            #   request.add_response_callback(modify_session)
            # this will cause session to be invalid
            request.add_response_callback(self._save, session=sess)

    @asyncio.coroutine
    def _save(self, request, session):
        """Save session.
        """
        if not session._changed:
            return
        session_id = yield from self._backend.save_session_data(session)
        self._sid_store.put_session_id(request, session_id)


def create_session_factory(session_id_store, backend_store, loop=None):
    """Creates new session factory.

    Create new session factory from two storage:
    session_id_store and backend_store.
    """
    return _SessionFactory(session_id_store=session_id_store,
                           backend_store=backend_store,
                           loop=loop)



import asyncio
import collections
import inspect
import json
import re
import aiohttp

from types import MethodType

from . import errors
from .handler import RESTRequestHandler


__all__ = [
    'RESTServer',
    ]


Entry = collections.namedtuple('Entry', 'regex method handler use_request'
                                        ' check_cors cors_options')


class RESTServer:

    DYN = re.compile(r'^\{[_a-zA-Z][_a-zA-Z0-9]*\}$')
    GOOD = r'[^{}/]+'
    PLAIN = re.compile('^'+GOOD+'$')

    METHODS = {'POST', 'GET', 'PUT', 'DELETE', 'PATCH', 'HEAD'}

    CORS_OPTIONS = {
        'allow-origin': '*',
        'allow-credentials': False,
        'allow-headers': None,
        'expose-headers': None,
        'max-age': 86400,
    }

    def __init__(self, *, hostname, session_factory=None,
                 enable_cors=False, loop=None, **kwargs):
        assert session_factory is None or callable(session_factory), \
            "session_factory must be None or callable (coroutine) function"
        if loop is None:
            loop = asyncio.get_event_loop()
        self._loop = loop
        super().__init__()
        self.hostname = hostname
        self.session_factory = session_factory
        self._enable_cors = enable_cors
        self._kwargs = kwargs
        self._urls = []

    def make_handler(self):
        return RESTRequestHandler(self, hostname=self.hostname,
                                  session_factory=self.session_factory,
                                  loop=self._loop,
                                  **self._kwargs)

    @property
    def cors_enabled(self):
        return self._enable_cors

    def add_url(self, method, path, handler, use_request=False,
                check_cors=True, cors_options={}):
        """XXX"""
        assert callable(handler), handler
        assert not set(cors_options) - set(self.CORS_OPTIONS), \
            'Got bad CORS options: {}'.format(
                set(cors_options) - set(self.CORS_OPTIONS))
        if isinstance(handler, MethodType):
            holder = handler.__func__
        else:
            holder = handler
        sig = holder.__signature__ = inspect.signature(holder)

        if use_request:
            if use_request is True:
                use_request = 'request'
            try:
                p = sig.parameters[use_request]
            except KeyError:
                raise TypeError('handler {!r} has no argument {}'
                                .format(handler, use_request))
            assert p.annotation is p.empty, ("handler's arg {} "
                                             "for request name "
                                             "should not have "
                                             "annotation").format(use_request)
        else:
            use_request = None
        assert path.startswith('/')
        assert callable(handler), handler
        method = method.upper()
        assert method in self.METHODS, method
        regexp = []
        for part in path.split('/'):
            if not part:
                continue
            if self.DYN.match(part):
                regexp.append('(?P<'+part[1:-1]+'>'+self.GOOD+')')
            elif self.PLAIN.match(part):
                regexp.append(part)
            else:
                raise ValueError("Invalid path '{}'['{}']".format(path, part))
        pattern = '/' + '/'.join(regexp)
        if path.endswith('/') and pattern != '/':
            pattern += '/'
        try:
            compiled = re.compile('^' + pattern + '$')
        except re.error:
            raise ValueError("Invalid path '{}'".format(path))
        cors_options = collections.ChainMap(cors_options, self.CORS_OPTIONS)
        allow_origin = cors_options.get('allow-origin')
        if allow_origin is not None:
            assert callable(allow_origin) \
                or isinstance(allow_origin, (collections.Sequence, str)), \
                "Invalid 'allow-origin' option {!r}".format(allow_origin)
        self._urls.append(Entry(compiled, method, handler, use_request,
                                check_cors, cors_options))

    @asyncio.coroutine
    def dispatch(self, request):
        path = request.path
        method = request.method
        allowed_methods = set()
        check_cors = False
        if method == 'OPTIONS' and self.cors_enabled:
            check_cors = True
            method = request.headers.get('ACCESS-CONTROL-REQUEST-METHOD')
            if not method:
                raise errors.RESTError(404, "Not Found")
        for entry in self._urls:
            match = entry.regex.match(path)
            if match is None:
                continue
            if entry.method != method:
                allowed_methods.add(entry.method)
            elif check_cors and entry.check_cors:
                headers = tuple(self._make_cors_headers(request,
                                                        entry.cors_options))
                raise errors.HttpCorsOptions(headers)
            else:
                break
        else:
            if allowed_methods:
                allow = ', '.join(sorted(allowed_methods))
                # add log
                raise errors.RESTError(405, "Not Allowed",
                                       json_body={'allowed_methods': allow},
                                       headers=(('Allow', allow),))
            else:
                # add log
                raise errors.RESTError(404, "Not Found")
        if self.cors_enabled and entry.check_cors:
            headers = tuple(self._make_cors_headers(request,
                                                    entry.cors_options))
            request.response.headers.extend(headers)

        handler = entry.handler
        sig = inspect.signature(handler)
        kwargs = match.groupdict()
        if entry.use_request:
            assert entry.use_request not in kwargs, (entry.use_request, kwargs)
            kwargs[entry.use_request] = request
        try:
            args, kwargs, ret_ann = self.construct_args(sig, kwargs)
            if asyncio.iscoroutinefunction(handler):
                ret = yield from handler(*args, **kwargs)
            else:
                ret = handler(*args, **kwargs)
            if ret_ann is not None:
                ret = ret_ann(ret)
        except aiohttp.HttpException as exc:
            raise
        except Exception as exc:
            # add log about error
            raise aiohttp.HttpErrorException(500,
                                             "Internal Server Error") from exc
        else:
            return json.dumps(ret)

    def construct_args(self, sig, kwargs):
        try:
            bargs = sig.bind(**kwargs)
        except TypeError:
            raise
        else:
            args = bargs.arguments
            marker = object()
            for name, param in sig.parameters.items():
                if param.annotation is param.empty:
                    continue
                val = args.get(name, marker)
                if val is marker:
                    continue    # Skip default value
                try:
                    args[name] = param.annotation(val)
                except (TypeError, ValueError) as exc:
                    raise ValueError(
                        'Invalid value for argument {!r}: {!r}'
                        .format(name, exc)) from exc
            if sig.return_annotation is not sig.empty:
                return bargs.args, bargs.kwargs, sig.return_annotation
            return bargs.args, bargs.kwargs, None

    def _make_cors_headers(self, request, cors_options):
        option = cors_options.get
        header = request.headers.get

        # TODO: implement real CORS check
        origin = header('ORIGIN')
        allow_origin = option('allow-origin')
        method = header('ACCESS-CONTROL-REQUEST-METHOD')

        # check bad request cases
        if origin is None or origin == '*' or not allow_origin:
            return
        if request.method == 'OPTIONS' and not method:
            return

        allow_creds = option('allow-credentials')
        if callable(allow_origin):
            allow_origin = allow_origin(request, cors_options)
        if isinstance(allow_origin, str):
            allow_origin = (allow_origin,)
        if '*' in allow_origin:
            if allow_creds:
                yield ('Access-Control-Allow-Origin', origin)
            else:
                yield ('Access-Control-Allow-Origin', '*')
        elif origin in allow_origin:
            yield ('Access-Control-Allow-Origin', origin)

        if method:
            yield ('Access-Control-Allow-Methods', method)

        allow_headers = option('allow-headers')
        if allow_headers:
            assert isinstance(allow_headers, (str, collections.Sequence))
            if not isinstance(allow_headers, str):
                allow_headers = ', '.join(allow_headers)
            yield ('Access-Control-Allow-Headers', allow_headers)
        if allow_creds:
            yield ('Access-Control-Allow-Credentials', allow_creds and 'true')



import aiohttp
import json


class RESTError(aiohttp.HttpException):
    """REST server error.

    Adds json_body to base aiohttp.HttpException.
    """

    def __init__(self, code, message='', json_body={}, headers=None):
        self.code = code
        self.message = message
        if json_body is None:
            self.body = None
        else:
            body = {'error': json_body,
                    'error_reason': message,
                    'error_code': code}
            self.body = json.dumps(body).encode('utf-8')
        self.headers = headers

    def write_response(self, response):
        if self.body is not None:
            response.add_headers(
                ('Content-Type', 'application/json; charset=utf-8'),
                ('Content-Length', str(len(self.body))),
                )
        else:
            response.add_headers(
                ('Content-Type', 'text/plain'),
                ('Content-Length', '0'),
                )
        if self.headers:
            response.add_headers(*self.headers)
        response.send_headers()

        if self.body is not None:
            response.write(self.body)
        response.write_eof()


class HttpCorsOptions(RESTError):
    """Http exception to handle CORS preflight requests.

    Raised in RESTServer.dispatch.
    """

    def __init__(self, headers):
        super().__init__(200, json_body=None, headers=headers)
